from flockai.PyCatascopia.probelib.ProcessProbe import ProcessProbe
from flockai.models.drones.autopilot_controlled_drone import AutopilotControlledDrone
from flockai.models.drones.keyboard_controller_drone import KeyboardControlledDrone
from flockai.utils.string_generator import StringGenerator
from flockai.models.devices.device_enums import EnableableDevice, NonEnableableDevice, MotorDevice
from flockai.utils.intensive_thread import IntensiveThread


class KeyboardMavic2DJI(KeyboardControlledDrone):
    """
    A Keyboard Controlled Mavic2DJI
    """

    def __init__(self, devices, probes=None):
        self.P_COMM = 8.4
        self.DJI_A3_FC = 8
        self.RASPBERRY_PI_4B_IDLE = 4
        self.RASPBERRY_PI_4B_ACTIVE = 8
        super().__init__(devices)
        self.probes = probes
        self._activate_probes()

    def _activate_probes(self):
        if self.probes:
            for key, probe in self.probes.items():
                probe.activate()
                print(f'{key} probe activated')

    def send_msg(self, msg, emitter_devices: list):
        """
        Sends a message from the attached emitter device
        :param emitter_devices: The emitter devices to send the message
        :param msg: The message to send
        :return: None
        """

        # total_time = 0
        # cpu_probe: ProcessProbe = self.probes['cpu']

        for emitter in emitter_devices:
            if emitter not in self.devices:
                print(f"The specified emitter device is not found: {emitter}")
                return
            f_emitter = self.devices[emitter]['device']
            # t1 = cpu_probe.probe_alive_time.val
            f_emitter.send(msg.encode('utf-8'))
            # t2 = cpu_probe.probe_alive_time.val
            # total_time += t2 - t1

        # return total_time
        return 0

    def receive_msgs(self, receiver_devices: list):
        """
        Receive messages on the receiver device
        :param receiver_devices: The receiver device name as a string
        :return: None
        """
        # cpu_probe: ProcessProbe = self.probes['cpu']
        messages = []

        for receiver in receiver_devices:
            if receiver not in self.devices:
                print(f"The specified receiver device is not found: {receiver}")
                return

            f_receiver = self.devices[receiver]['device']
            while f_receiver.getQueueLength() > 0:
                # t1 = cpu_probe.probe_alive_time.val
                message_received = f_receiver.getData().decode('utf-8')
                messages.append(message_received)
                f_receiver.nextPacket()
                # t2 = cpu_probe.probe_alive_time.val
                # yield message_received, t2 - t1

        return messages, 0

    def blink_led_lights(self, led_devices: list):
        """
        Blink led lights every 2 seconds
        :param time: The simulation time. ie self.getTime()
        :return:
        """
        f_led_devices = [self.devices[led]['device'] for led in led_devices]
        led_on = (self.getTime() // 1) % 2 == 1

        for f_led in f_led_devices:
            f_led.set(led_on)
            led_on = not led_on

    def run(self):
        # Wait a second before starting
        while self.step(self.basic_time_step) != -1:
            if self.getTime() > 1:
                break

        led_devices = []
        emitter_devices = []
        receiver_devices = []
        for device_name in self.devices:
            if self.devices[device_name]['type'] == NonEnableableDevice.LED:
                led_devices.append(device_name)
            if self.devices[device_name]['type'] == NonEnableableDevice.EMITTER:
                emitter_devices.append(device_name)
            if self.devices[device_name]['type'] == EnableableDevice.RECEIVER:
                receiver_devices.append(device_name)

        # self._activate_probes()
        cpu_probe: ProcessProbe = self.probes['cpu']
        total_energy = 0
        net_transaction_time = 0

        while self.step(self.basic_time_step) != -1:
            # Blink lights
            self.blink_led_lights(led_devices)
            # Fly drone
            self.actuate()

            # Get battery data
            # print(self.batterySensorGetValue())

            # Do an intensive task
            # t = IntensiveThread()
            # t.run()

            # Send messages
            print('before tt')
            transmit_time = self.send_msg(StringGenerator.get_random_message(4), emitter_devices)

            # Receive messages
            print('before rm')
            received_messages, receive_time = self.receive_msgs(receiver_devices)

            print('before calculations')
            flight_time = cpu_probe.probe_alive_time.val
            cpu_time = cpu_probe.cpu_time.val
            io_time = 0 if cpu_probe.io_time.val is None else cpu_probe.io_time.val
            idle_time = flight_time - (cpu_time + io_time)

            total_energy += flight_time * self.DJI_A3_FC
            print(total_energy)


class AutopilotMavic2DJI(AutopilotControlledDrone):
    def __init__(self, devices):
        super().__init__(devices)

    def run(self):
        # Wait a second before starting
        while self.step(self.basic_time_step) != -1:
            if self.getTime() > 1:
                break

        while self.step(self.basic_time_step) != -1:
            # For now just actuate
            self.actuate()



from flockai.interfaces.energy import IEnergy


class MotorEnergy(IEnergy):
    """
    Energy used to operate motors. Ie. Electrical Motors, Propellers...
    """
    def __init__(self):
        self.take_off_energy = 0  # energy attributed to the drone lifted in the air till it enters a hovering state
        self.hover_energy = 0  # energy attributed to keeping the drone in the air and is dependent to the air density and weight of the drone
        self.move_energy = 0  # the kinetic energy required to set a drone in motion from a hovering state

    def calculate(self):
        return self.take_off_energy + \
               self.hover_energy + \
               self.move_energy

    def set(self, *args, **kwargs):
        pass


class CommunicationEnergy(IEnergy):
    """
    Energy used to operate communication devices
    """

    def __init__(self):
        self.p_transmit = 0  # energy attributed to transmit data
        self.t_transmit = 0  # time attributed to transmit data
        self.p_receive = 0   # energy attributed to receive data
        self.t_receive = 0   # energy attributed to receive data
        self.p_idle = 0      # energy attributed when idling
        self.t_idle = 0      # idle time

    def calculate(self):
        return self.p_transmit * self.t_transmit + \
               self.p_receive * self.t_receive + \
               self.p_idle * self.t_idle

    def set(self, p_transmit=0, p_receive=0, p_idle=0):
        self.p_transmit = p_transmit
        self.p_receive = p_receive
        self.p_idle = p_idle


class ProcessingEnergy(IEnergy):
    """
    Energy used for processing
    """
    def __init__(self):
        self.p_active = 0      # processor power in compute state
        self.t_active = 0      # cpu time
        self.p_io = 0       # processor power in I/O state
        self.t_io = 0       # I/O time
        self.p_peripheral = 0  # processor power driving peripheral sensing devices
        self.t_peripheral = 0  # peripheral usage time
        self.p_idle = 0     # processor power when idling
        self.t_idle = 0     # idle time

    def calculate(self):
        return self.p_active * self.t_active + \
               self.p_io * self.t_io + \
               self.p_peripheral * self.t_peripheral + \
               self.p_idle * self.t_idle

    def set(self, p_active=0, p_io=0, p_peripheral=0, p_idle=0):
        self.p_active = p_active
        self.p_io = p_io
        self.p_peripheral = p_peripheral
        self.p_idle = p_idle


class Energy:
    def __init__(self):
        self.motor_energy = MotorEnergy()
        self.communication_energy = CommunicationEnergy()
        self.processing_energy = ProcessingEnergy()




import abc

from flockai.interfaces.drone import IDrone


class AutopilotControlledDrone(IDrone, abc.ABC):
    def __init__(self, devices):
        super().__init__(devices)

    def set_flight_plan(self, plan):
        pass

    def update_flight_plan(self, plan):
        pass

    def abort_flight_plan(self):
        pass

    def get_flight_plan(self):
        pass

    def get_input(self):
        # Transform the keyboard input to disturbances on the stabilization algorithm.
        roll_disturbance = 0
        pitch_disturbance = 2
        yaw_disturbance = 0

        return roll_disturbance, pitch_disturbance, yaw_disturbance



from enum import IntEnum, auto


class Devices:
    def __init__(self, enableable_devices, non_enableable_devices, motor_devices):
        self.enableable_devices = enableable_devices
        self.non_enableable_devices = non_enableable_devices
        self.motor_devices = motor_devices


class EnableableDevice(IntEnum):
    RECEIVER = 0
    CAMERA = auto()
    KEYBOARD = auto()
    BATTERY_SENSOR = auto()
    GPS = auto()
    COMPASS = auto()
    INERTIAL_UNIT = auto()
    GYRO = auto()


class NonEnableableDevice(IntEnum):
    LED = 100
    EMITTER = auto()


class MotorDevice(IntEnum):
    CAMERA = 200
    PROPELLER = auto()


class AircraftAxis(IntEnum):
    ROLL = 300
    PITCH = auto()
    YAW = auto()


class Relative2DPosition:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        x = 'front' if self.x == 1 else 'rear'
        y = 'left' if self.y == 1 else 'right'
        return f"{x} {y}"



import abc
from flockai.interfaces.robot import IRobot
import math

from flockai.models.devices.device_enums import EnableableDevice, MotorDevice, AircraftAxis, Relative2DPosition, \
    NonEnableableDevice
from flockai.models.energy.energy import Energy
from flockai.utils.graphics import Graphics


class IDrone(IRobot, abc.ABC):
    """
    Define all actions needed for flying a drone and declare the get_input() abstract method
    """

    def __init__(self, devices):
        super().__init__()
        self.name = self.getName()
        self.energy_model = Energy()
        self.basic_time_step = int(self.getBasicTimeStep())
        self.devices = self._attach_and_enable_devices(devices.enableable_devices, devices.non_enableable_devices)
        self.motors = self._attach_and_enable_motors(devices.motor_devices)
        self._cross_check_devices()
        self._cross_check_motors()
        self._set_constants()
        self._set_variables()

    def _attach_and_enable_devices(self, en_devices, nen_devices):
        """
        Define all the required devices and enable them
        :return:
        """
        e_devices = {}
        for device, name in en_devices:
            if EnableableDevice(device) == EnableableDevice.KEYBOARD:
                e_devices['keyboard'] = {'type': device, 'device': self.keyboard}
                e_devices['keyboard']['device'].enable(self.basic_time_step)
            elif EnableableDevice(device) == EnableableDevice.BATTERY_SENSOR:
                self.batterySensorEnable(self.basic_time_step)
            elif name is not None:
                e_devices[name] = {'type': device, 'device': self.getDevice(name)}
                e_devices[name]['device'].enable(self.basic_time_step)

        ne_devices = {}
        for device, name in nen_devices:
            if name is not None:
                ne_devices[name] = {'type': device, 'device': self.getDevice(name)}

        return {**e_devices, **ne_devices}

    def _attach_and_enable_motors(self, motor_devices):
        """
        Define all the required motors and enable them
        :return:
        """
        m_devices = {}
        for device, name, *args in motor_devices:
            if MotorDevice(device) == MotorDevice.CAMERA:
                m_devices[name] = {'type': device, 'motor': self.getDevice(name), 'axis': args[0]}
            elif MotorDevice(device) == MotorDevice.PROPELLER:
                m_devices[name] = {'type': device, 'motor': self.getDevice(name),
                                   'relative_position': args[0]}
                m_devices[name]['motor'].setPosition(float('inf'))
                m_devices[name]['motor'].setVelocity(1.0)

        return m_devices

    def _cross_check_devices(self):
        required_devices = {
            EnableableDevice.INERTIAL_UNIT: True,
            EnableableDevice.GPS: True,
            EnableableDevice.GYRO: True,
        }
        for device in self.devices.keys():
            f_device = self.devices[device]['device']
            f_type = self.devices[device]['type']
            if f_type in required_devices:
                if EnableableDevice(f_type) == EnableableDevice.INERTIAL_UNIT:
                    self.imu = f_device
                if EnableableDevice(f_type) == EnableableDevice.GPS:
                    self.gps = f_device
                if EnableableDevice(f_type) == EnableableDevice.GYRO:
                    self.gyro = f_device
                required_devices[f_type] = False

        for key, value in required_devices.items():
            if value:
                raise NotImplementedError(f"The drone needs {EnableableDevice(key).name} device to operate")

    def _cross_check_motors(self):
        front_left = repr(Relative2DPosition(1, -1))
        front_right = repr(Relative2DPosition(1, 1))
        rear_left = repr(Relative2DPosition(-1, -1))
        rear_right = repr(Relative2DPosition(-1, 1))

        required_motors = {
            MotorDevice.CAMERA: {
                AircraftAxis.ROLL: True,
                AircraftAxis.PITCH: True,
                AircraftAxis.YAW: True
            },
            MotorDevice.PROPELLER: {
                front_left: True,
                front_right: True,
                rear_left: True,
                rear_right: True,
            }
        }

        for motor in self.motors.keys():
            f_motor = self.motors[motor]['motor']
            f_type = self.motors[motor]['type']
            if f_type in required_motors:
                arg = None
                if MotorDevice(f_type) == MotorDevice.CAMERA:
                    f_axis = AircraftAxis(self.motors[motor]['axis'])
                    if AircraftAxis(f_axis) == AircraftAxis.ROLL:
                        self.camera_roll_motor = f_motor
                    if AircraftAxis(f_axis) == AircraftAxis.PITCH:
                        self.camera_pitch_motor = f_motor
                    if AircraftAxis(f_axis) == AircraftAxis.YAW:
                        self.camera_yaw_motor = f_motor
                    arg = f_axis
                elif MotorDevice(f_type) == MotorDevice.PROPELLER:
                    f_relative_position = repr(self.motors[motor]['relative_position'])
                    if f_relative_position == repr(Relative2DPosition(1, -1)):
                        self.front_left_motor = f_motor
                    if f_relative_position == repr(Relative2DPosition(1, 1)):
                        self.front_right_motor = f_motor
                    if f_relative_position == repr(Relative2DPosition(-1, -1)):
                        self.rear_left_motor = f_motor
                    if f_relative_position == repr(Relative2DPosition(-1, 1)):
                        self.rear_right_motor = f_motor
                    arg = f_relative_position
                required_motors[f_type][arg] = False

        for motor_key in required_motors.keys():
            for key, value in required_motors[motor_key].items():
                if value:
                    motor_device = MotorDevice(motor_key).name
                    if MotorDevice(motor_key) == MotorDevice.CAMERA:
                        motor_axis = AircraftAxis(key).name
                        if value:
                            raise NotImplementedError(f"The drone needs {motor_device} {motor_axis} motor to operate")
                    if MotorDevice(motor_key) == MotorDevice.PROPELLER:
                        if value:
                            raise NotImplementedError(f"The drone needs {key} {motor_device} to operate")

    def _set_variables(self):
        """
        Set variables needed for controlling the drone on air
        :return:
        """
        self.target_altitude = 1.0  # The target altitude. Can be modified

    def _set_constants(self):
        """
        Set constants needed for controlling the drone on air
        :return: None
        """
        self.K_VERTICAL_THRUST = 68.5  # this thrust lifts the drone
        self.K_VERTICAL_OFFSET = 0.6  # Vertical offset where the robot actually targets to stabilize itself
        self.K_VERTICAL_P = 3.0  # P constant of the vertical PID
        self.K_ROLL_P = 50.0  # P constant of the roll PID
        self.K_PITCH_P = 30.0  # P constant of the pitch PID
        self.K_YAW_P = 20.0  # P constant of the yaw PID

        self.energy_model.processing_energy.set(p_active=self.RASPBERRY_PI_4B_ACTIVE,
                                                p_io=self.RASPBERRY_PI_4B_IDLE,
                                                p_idle=self.RASPBERRY_PI_4B_IDLE,
                                                p_peripheral=0)

        self.energy_model.communication_energy.set(p_transmit=self.P_COMM,
                                                   p_receive=self.P_COMM,
                                                   p_idle=self.P_COMM / 2)

    def actuate(self):
        """
        Fly the drone by stabilizing it based on camera motor sensors and input
        :return: None
        """
        roll, pitch, yaw = self.imu.getRollPitchYaw()
        roll += math.pi / 2

        altitude = self.gps.getValues()[1]
        roll_acceleration, pitch_acceleration, yaw_acceleration = self.gyro.getValues()

        # Set camera motors based on current gyro data
        self.camera_roll_motor.setPosition(-0.115 * roll_acceleration)
        self.camera_pitch_motor.setPosition(-0.1 * pitch_acceleration)

        # Get input
        roll_disturbance, pitch_disturbance, yaw_disturbance = self.get_input()

        roll_input = self.K_ROLL_P * Graphics.clamp(roll, -1.0, 1.0) + roll_acceleration + roll_disturbance
        pitch_input = self.K_PITCH_P * Graphics.clamp(pitch, -1.0, 1.0) - pitch_acceleration + pitch_disturbance
        yaw_input = yaw_disturbance
        clamped_difference_altitude = Graphics.clamp(self.target_altitude - altitude + self.K_VERTICAL_OFFSET, -1.0,
                                                     1.0)
        vertical_input = self.K_VERTICAL_P * pow(clamped_difference_altitude, 3.0)

        # Actuate based on inputs
        front_left_motor_input = self.K_VERTICAL_THRUST + vertical_input - roll_input - pitch_input + yaw_input
        front_right_motor_input = self.K_VERTICAL_THRUST + vertical_input + roll_input - pitch_input - yaw_input
        rear_left_motor_input = self.K_VERTICAL_THRUST + vertical_input - roll_input + pitch_input - yaw_input
        rear_right_motor_input = self.K_VERTICAL_THRUST + vertical_input + roll_input + pitch_input + yaw_input

        self.front_left_motor.setVelocity(front_left_motor_input)
        self.front_right_motor.setVelocity(-front_right_motor_input)
        self.rear_left_motor.setVelocity(-rear_left_motor_input)
        self.rear_right_motor.setVelocity(rear_right_motor_input)

    @abc.abstractmethod
    def get_input(self) -> tuple:
        """
        Get the input for actuating the drone
        :return: A tuple with pitch, roll, yaw disturbance to actuate the drone
        """
        raise NotImplementedError



import time


class Metric(object):

    def __init__(self, name, units, desc):
        self.name = name
        self.units = units
        self.desc = desc
        self.val = None
        self.timestamp = None
        self.group = None

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    def get_units(self):
        return self.units

    def set_units(self, units):
        self.units = units

    def get_desc(self):
        return self.desc

    def set_desc(self, desc):
        self.desc

    def get_timestamp(self):
        return self.timestamp

    def set_timestamp(self, timestamp):
        self.timestamp

    def get_val(self):
        return self.val

    def set_val(self, val):
        self.val = val

    def get_group(self):
        return self.group

    def set_group(self, group):
        self.group = group

    def to_dict(self):
        d = dict()
        d['name'] = self.name
        d['units'] = self.units
        d['desc'] = self.desc
        d['timestamp'] = self.timestamp
        d['val'] = self.val
        d['group'] = self.group
        return d

    def __str__(self):
        return str(self.to_dict())


class SimpleMetric(Metric):

    def __init__(self, name, units, desc, minVal = float('-inf'), maxVal = float('inf'), higherIsBetter=True):
        super(SimpleMetric, self).__init__(name, units, desc)
        self.higherIsBetter = higherIsBetter
        self.minVal = minVal
        self.maxVal = maxVal

    def get_higherisbetter(self):
        return self.higherIsBetter

    def set_higherisbetter(self, higherIsBetter):
        self.higherIsBetter = higherIsBetter

    def get_minval(self):
        return self.minVal

    def set_minval(self, minVal):
        self.minVal = minVal

    def get_maxval(self):
        return self.maxVal

    def set_maxval(self, maxVal):
        self.maxVal = maxVal

    def set_val(self, val):
        self.timestamp = int(time.time())
        self.val = val

    def __str__(self):
        m = super().to_dict()
        m['higherIsBetter'] = self.higherIsBetter
        m['minVal'] = self.minVal
        m['maxVal'] = self.maxVal
        return str(m)

class CounterMetric(SimpleMetric):
    def __init__(self, name, units, desc, minVal = 0, maxVal = float('inf'), higherIsBetter=True, step=1, reset=True):
        super(CounterMetric, self).__init__(name, units, desc, minVal, maxVal, higherIsBetter)
        self.step = step
        self.counter = minVal
        self.reset = reset

    def inc(self):
        self._inc(self.step)

    def _inc(self, step):
        if isinstance(step, int):
            self.counter += step
            if self.counter > self.maxVal:
                if self.reset:
                    self.counter = self.minVal
                else:
                    raise CatascopiaMetricValueException('CounterMetric ' + self.name + ' counter max value overflow')
            self.set_val(self.counter)
        else:
            raise CatascopiaMetricValueException('CounterMetric ' + self.name + ' step ' + step + ' is not an integer')


class DiffMetric(SimpleMetric):

    def __init__(self, name, units='%', desc='a diff metric', minVal = float('-inf'), maxVal = float('inf'), higherIsBetter=True):
        super(DiffMetric, self).__init__(name, units, desc, minVal, maxVal, higherIsBetter)
        self.prev = 0
        self.cur = 0
        self.diff = 0

    def update(self, val):
        self.prev = self.cur
        self.cur = val
        try:
            self.diff = (self.cur - self.prev)/self.prev
        except TypeError:
            raise CatascopiaMetricValueException('DiffMetric ' + self.name + ' value type ' + str(type(val))
                                                 + ' incompatible, expected numeric value')
        except ZeroDivisionError:
            self.diff = 0
        self.set_val(self.diff * 100)

# TODO implement TimerMetric class
class TimerMetric(Metric):
    pass


class CatascopiaMetricValueException(Exception):
    pass


import psutil
from time import time

from flockai.PyCatascopia.Probe import Probe
from flockai.PyCatascopia.Metrics import SimpleMetric

# use pid param if the monitored process different from the current process

class ProcessProbe(Probe):

    __PROC_TIMESTEP__ = 0.2

    def __init__(self, name = "ProcessProbe", periodicity = 5, pid = None):
        super(ProcessProbe, self).__init__(name, periodicity)

        self.cpu_percent = SimpleMetric('cpu_percent', '%', 'process-level cpu utilization', minVal=0, higherIsBetter=False)
        self.cpu_time = SimpleMetric('cpu_time', 's', 'process-level cpu time', minVal=0, higherIsBetter=False)
        self.io_time = SimpleMetric('io_time', 's', 'process-level io time (linux-only)', minVal=0, higherIsBetter=False)
        self.alive_time = SimpleMetric('alive_time', 's', 'time process is alive', minVal=0, higherIsBetter=False)
        self.probe_alive_time = SimpleMetric('probe_alive_time', 's', 'time probe is alive', minVal=0, higherIsBetter=False)

        self.add_metric(self.cpu_percent)
        self.add_metric(self.cpu_time)
        self.add_metric(self.io_time)
        self.add_metric(self.alive_time)
        self.add_metric(self.probe_alive_time)

        self.proc = psutil.Process(pid)
        self.col_start = time()

    def get_desc(self):
        return "ProcessProbe collects process-level utilization metrics..."

    def collect(self):
        self.cpu_percent.set_val(self.proc.cpu_percent(interval=ProcessProbe.__PROC_TIMESTEP__))
        # user+sys including child threads
        ct = self.proc.cpu_times()
        self.cpu_time.set_val(sum(ct[:4]))
        try:
           self.io_time.set_val(ct[4])
        except IndexError:
            pass # not a Linux OS, so no io_time counter

        self.alive_time.set_val(time() - self.proc.create_time())
        self.probe_alive_time.set_val(time() - self.col_start)

def main():
    p = ProcessProbe()
    p.set_debugmode(True)
    p.set_logging()
    p.activate()

if __name__ == "__main__":
    main()